import astropy.units as u
import dacite
import yaml

from dataclasses import dataclass
import os
import pathlib
import sys
from typing import Union, Dict


DEFAULT_LIBRARY_PATH = pathlib.Path(__file__).parent / 'data' / 'astrolabel.yml'


@dataclass
class AstroLabel:
    symbol: str
    unit: Union[str, None] = None
    description: Union[str, None] = None


@dataclass
class LabelLibrary:
    formats: Dict[str, str]
    labels: Dict[str, AstroLabel]

    def __post_init__(self):
        self._library_path: pathlib.Path | None = None

    def library_fname(self):
        return self._library_path

    def info(self, output=None):
        if output is None:
            output = sys.stdout

        library_summary = []
        max_key_len = max(map(len, self.labels.keys()))
        for label_key, label_data in self.labels.items():
            library_summary.append(f'{label_key:>{max_key_len}}: {label_data.description}')

        output.write("\n".join(library_summary))
        output.write("\n")
        output.flush()

    @staticmethod
    def _get_library_path() -> pathlib.Path:
        # search for a library in the working directory
        library_path = pathlib.Path() / "astrolabel.yml"
        if library_path.exists():
            return library_path

        # use the path stored in the environment variable - if not set, use the path to the default library
        library_path = os.environ.get("ASTROLABEL", default=DEFAULT_LIBRARY_PATH)
        return pathlib.Path(library_path)

    @classmethod
    def read(cls, filename: Union[str, pathlib.Path, None] = None):
        if filename is None:
            library_path = cls._get_library_path()
        else:
            library_path = pathlib.Path(filename)

        library_path = library_path.resolve()
        if not library_path.is_file():
            raise FileNotFoundError(f"File \"{library_path}\" does not exist")

        with open(library_path, "r") as label_library:
            label_data = yaml.safe_load(label_library)

        # create the LabelLibrary object
        ll = dacite.from_dict(data_class=cls, data=label_data, config=dacite.Config(strict=True))

        # store the path to the label library
        ll._library_path = library_path

        return ll

    @staticmethod
    def _replace(label, key, value):
        i = label.index(key)
        if label[:i].count('$') % 2 == 1:
            value = value[1:-1]  # strip dollar signs
        return label.replace(key, value)

    def get_label(self, name: str, fmt: str = 'default'):
        symbol = f'${self.labels[name].symbol}$'  # treat the symbol as math text
        unit_plain = self.labels[name].unit

        label = self.formats[fmt + '_u'] if unit_plain else self.formats[fmt]
        label = self._replace(label, '__symbol__', symbol)

        if unit_plain:
            unit = u.Unit(unit_plain).to_string('latex_inline')
            label = self._replace(label, '__unit__', unit)

        return label



from astrolabel import LabelLibrary, AstroLabel, DEFAULT_LIBRARY_PATH
import pytest

import shutil


@pytest.fixture(autouse=True)
def set_env(monkeypatch):
    monkeypatch.delenv("ASTROLABEL", raising=False)


@pytest.fixture
def ll() -> LabelLibrary:
    ll = LabelLibrary.read()
    assert ll.library_fname() == DEFAULT_LIBRARY_PATH

    return ll


@pytest.fixture
def tmp_library_path(tmp_path):
    library_path = tmp_path / "astrolabel.yml"
    shutil.copy(DEFAULT_LIBRARY_PATH, library_path)

    return library_path


def test_info(capsys):
    ll = LabelLibrary(formats={}, labels={'z': AstroLabel(symbol='z', description='Redshift'),
                                         'fesc': AstroLabel(symbol='f_{\\text{esc}}', description='Escape fraction')})
    ll.info()

    captured = capsys.readouterr()
    assert captured.out == "   z: Redshift\nfesc: Escape fraction\n"


def test_read_from_file(tmp_library_path):
    ll = LabelLibrary.read(tmp_library_path)

    assert ll.library_fname() == tmp_library_path


def test_read_from_env_variable(tmp_library_path, monkeypatch):
    monkeypatch.setenv("ASTROLABEL", str(tmp_library_path))
    ll = LabelLibrary.read()

    assert ll.library_fname() == tmp_library_path


def test_read_from_workdir(tmp_library_path, monkeypatch):
    monkeypatch.chdir(tmp_library_path.parent)
    ll = LabelLibrary.read()

    assert ll.library_fname() == tmp_library_path


def test_read_not_found():
    with pytest.raises(FileNotFoundError, match="File \".*\" does not exist"):
        LabelLibrary.read("non_existing_file.yml")


def test_without_unit(ll):
    assert ll.get_label('z') == '$z$'


def test_with_unit(ll):
    assert ll.get_label('sfr') == r'$\mathrm{SFR}$ [$\mathrm{M_{\odot}\,yr^{-1}}$]'


def test_log_without_unit(ll):
    assert ll.get_label('z', fmt='log') == r'$\log_{10}\,z$'


def test_log_with_unit(ll):
    assert ll.get_label('sfr', fmt='log') == (r'$\log_{10}\,\left(\mathrm{SFR} / \mathrm{M_{\odot}\,'
                                              r'yr^{-1}}\right)$')
